=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2006-06-27 00:58:18 +0000
+++ bzrlib/branch.py	2006-06-30 19:58:02 +0000
@@ -995,12 +995,20 @@
     def lock_write(self):
         # TODO: test for failed two phase locks. This is known broken.
         self.control_files.lock_write()
-        self.repository.lock_write()
+        try:
+            self.repository.lock_write()
+        except:
+            self.control_files.unlock()
+            raise
 
     def lock_read(self):
         # TODO: test for failed two phase locks. This is known broken.
         self.control_files.lock_read()
-        self.repository.lock_read()
+        try:
+            self.repository.lock_read()
+        except:
+            self.control_files.unlock()
+            raise
 
     def unlock(self):
         # TODO: test for failed two phase locks. This is known broken.

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-06-20 05:32:16 +0000
+++ bzrlib/errors.py	2006-06-30 19:58:02 +0000
@@ -451,6 +451,10 @@
         self.lock = lock
 
 
+class TestPreventLocking(LockError):
+    """A test exception for forcing locking failure: %(message)s"""
+
+
 class PointlessCommit(BzrNewError):
     """No changes to commit"""
 

=== modified file 'bzrlib/lock.py'
--- bzrlib/lock.py	2006-06-20 03:57:11 +0000
+++ bzrlib/lock.py	2006-06-30 19:58:02 +0000
@@ -39,7 +39,66 @@
 import sys
 
 from bzrlib.trace import mutter
-from bzrlib.errors import LockError
+from bzrlib.errors import LockError, TestPreventLocking
+
+
+class LockWrapper(object):
+    """A wrapper which lets us set locking ability.
+
+    This also lets us record what objects were locked in what order,
+    to ensure that locking happens correctly.
+    """
+
+    def __init__(self, sequence, other, other_id):
+        """Wrap a locking policy around a given object.
+
+        :param sequence: A list object where we should record actions
+        :param other: The object to control policy on
+        :param other_id: Something to identify the object by
+        """
+        self.__dict__['_sequence'] = sequence
+        self.__dict__['_other'] = other
+        self.__dict__['_other_id'] = other_id
+        self.__dict__['_allow_write'] = True
+        self.__dict__['_allow_read'] = True
+        self.__dict__['_allow_unlock'] = True
+
+    def __getattr__(self, attr):
+        return getattr(self._other, attr)
+
+    def __setattr__(self, attr, val):
+        return setattr(self._other, attr, val)
+
+    def lock_read(self):
+        self._sequence.append((self._other_id, 'lr', self._allow_read))
+        if self._allow_read:
+            return self._other.lock_read()
+        raise TestPreventLocking('lock_read disabled')
+
+    def lock_write(self):
+        self._sequence.append((self._other_id, 'lw', self._allow_write))
+        if self._allow_write:
+            return self._other.lock_write()
+        raise TestPreventLocking('lock_write disabled')
+
+    def unlock(self):
+        self._sequence.append((self._other_id, 'ul', self._allow_unlock))
+        if self._allow_unlock:
+            return self._other.unlock()
+        raise TestPreventLocking('unlock disabled')
+
+    def disable_lock_read(self):
+        """Make a lock_read call fail"""
+        self.__dict__['_allow_read'] = False
+
+    def disable_unlock(self):
+        """Make an unlock call fail"""
+        self.__dict__['_allow_unlock'] = False
+
+    def disable_lock_write(self):
+        """Make a lock_write call fail"""
+        self.__dict__['_allow_write'] = False
+
 
 class _base_Lock(object):
     def _open(self, filename, filemode):
@@ -65,10 +124,6 @@
     def unlock(self):
         raise NotImplementedError()
 
-        
-
-
-
 
 ############################################################
 # msvcrt locks
@@ -141,7 +196,6 @@
                     raise LockError(e)
 
 
-
         class _w32c_ReadLock(_w32c_FileLock):
             def __init__(self, filename):
                 _w32c_FileLock._lock(self, filename, 'rb',
@@ -153,7 +207,6 @@
                                      LOCK_EX + LOCK_NB)
 
 
-
         WriteLock = _w32c_WriteLock
         ReadLock = _w32c_ReadLock
 
@@ -188,7 +241,6 @@
                     _msvc_lock(self._open(filename, 'wb'), self.LOCK_EX)
 
 
-
             def _msvc_lock(f, flags):
                 try:
                     # Unfortunately, msvcrt.LK_RLCK is equivalent to msvcrt.LK_LOCK
@@ -244,7 +296,6 @@
                     raise LockError(e)
 
 
-
             WriteLock = _msvc_WriteLock
             ReadLock = _msvc_ReadLock
         except ImportError:

=== modified file 'bzrlib/tests/branch_implementations/test_branch.py'
--- bzrlib/tests/branch_implementations/test_branch.py	2006-06-22 15:42:26 +0000
+++ bzrlib/tests/branch_implementations/test_branch.py	2006-06-30 19:58:02 +0000
@@ -27,7 +27,9 @@
                            NoSuchFile,
                            UninitializableFormat,
                            NotBranchError,
+                           TestPreventLocking,
                            )
+from bzrlib.lock import LockWrapper
 from bzrlib.osutils import getcwd
 from bzrlib.tests import TestCase, TestCaseWithTransport, TestSkipped
 from bzrlib.tests.bzrdir_implementations.test_bzrdir import TestCaseWithBzrDir
@@ -561,3 +563,154 @@
             return
         self.assertEqual(self.branch_format,
                          branch.BranchFormat.find_format(opened_control))
+
+
+class TestBranchLocking(TestCaseWithBranch):
+
+    def get_instrumented_branch(self):
+        """Get a Branch object which has been instrumented"""
+        # TODO: jam 20060630 It may be that not all formats have a 
+        # 'control_files' member. So we should fail gracefully if
+        # not there. But assuming it has them lets us test the exact 
+        # lock/unlock order.
+        self.locks = []
+        b = LockWrapper(self.locks, self.get_branch(), 'b')
+        b.repository = LockWrapper(self.locks, b.repository, 'r')
+        bcf = b.control_files
+        rcf = b.repository.control_files
+
+        # Look out for branch types that reuse their control files
+        self.combined_control = bcf is rcf
+
+        b.control_files = LockWrapper(self.locks, b.control_files, 'bc')
+        b.repository.control_files = \
+            LockWrapper(self.locks, b.repository.control_files, 'rc')
+        return b
+
+    def test_01_lock_read(self):
+        # Test that locking occurs in the correct order
+        b = self.get_instrumented_branch()
+
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+        b.lock_read()
+        try:
+            self.assertTrue(b.is_locked())
+            self.assertTrue(b.repository.is_locked())
+        finally:
+            b.unlock()
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lr', True), ('bc', 'lr', True),
+                          ('r', 'lr', True), ('rc', 'lr', True),
+                          ('b', 'ul', True), ('r', 'ul', True),
+                          ('rc', 'ul', True), ('bc', 'ul', True),
+                         ], self.locks)
+
+    def test_02_lock_write(self):
+        # Test that locking occurs in the correct order
+        b = self.get_instrumented_branch()
+
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+        b.lock_write()
+        try:
+            self.assertTrue(b.is_locked())
+            self.assertTrue(b.repository.is_locked())
+        finally:
+            b.unlock()
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lw', True), ('bc', 'lw', True),
+                          ('r', 'lw', True), ('rc', 'lw', True),
+                          ('b', 'ul', True), ('r', 'ul', True),
+                          ('rc', 'ul', True), ('bc', 'ul', True),
+                         ], self.locks)
+
+    def test_03_lock_fail_unlock_repo(self):
+        # Make sure branch.unlock() is called, even if there is a
+        # failure while unlocking the repository.
+        b = self.get_instrumented_branch()
+        b.repository.disable_unlock()
+
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+        b.lock_write()
+        try:
+            self.assertTrue(b.is_locked())
+            self.assertTrue(b.repository.is_locked())
+            self.assertRaises(TestPreventLocking, b.unlock)
+            if self.combined_control:
+                self.assertTrue(b.is_locked())
+            else:
+                self.assertFalse(b.is_locked())
+            self.assertTrue(b.repository.is_locked())
+
+            # We unlock the branch control files, even if 
+            # we fail to unlock the repository
+            self.assertEqual([('b', 'lw', True), ('bc', 'lw', True),
+                              ('r', 'lw', True), ('rc', 'lw', True),
+                              ('b', 'ul', True), ('r', 'ul', False), 
+                              ('bc', 'ul', True),
+                             ], self.locks)
+
+        finally:
+            # For cleanup purposes, make sure we are unlocked
+            b.repository._other.unlock()
+
+    def test_04_lock_read_fail_repo(self):
+        # Test that the branch is unlocked if it cannot lock the repository
+        b = self.get_instrumented_branch()
+        b.repository.disable_lock_read()
+
+        self.assertRaises(TestPreventLocking, b.lock_read)
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lr', True), ('bc', 'lr', True),
+                          ('r', 'lr', False), 
+                          ('bc', 'ul', True),
+                         ], self.locks)
+
+    def test_05_lock_write_fail_repo(self):
+        # Test that the branch is unlocked if it cannot lock the repository
+        b = self.get_instrumented_branch()
+        b.repository.disable_lock_write()
+
+        self.assertRaises(TestPreventLocking, b.lock_write)
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lw', True), ('bc', 'lw', True),
+                          ('r', 'lw', False), 
+                          ('bc', 'ul', True),
+                         ], self.locks)
+
+    def test_06_lock_read_fail_control(self):
+        # We shouldn't try to lock the repo if we can't lock the branch
+        b = self.get_instrumented_branch()
+        b.control_files.disable_lock_read()
+
+        self.assertRaises(TestPreventLocking, b.lock_read)
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lr', True), ('bc', 'lr', False),
+                         ], self.locks)
+
+    def test_07_lock_write_fail_control(self):
+        # We shouldn't try to lock the repo if we can't lock the branch
+        b = self.get_instrumented_branch()
+        b.control_files.disable_lock_write()
+
+        self.assertRaises(TestPreventLocking, b.lock_write)
+        self.assertFalse(b.is_locked())
+        self.assertFalse(b.repository.is_locked())
+
+        self.assertEqual([('b', 'lw', True), ('bc', 'lw', False),
+                         ], self.locks)
+
+
+

